from typing import List

from lucyfer.settings import lucyfer_settings


def escape_quotes(values: List[str]) -> List[str]:
    try:
        values = [v.replace("'", "\\'").replace('"', '\\"') for v in values]
    except (AttributeError, TypeError):  # if value is not str
        pass
    return values


def ignore_empty_values(values: List[str]) -> List[str]:
    return [v for v in values if v not in lucyfer_settings.EMPTY_SUGGESTIONS_VALUES]


def custom_sorted(values: List[str]) -> List[str]:
    return sorted(values)



from typing import List, Optional, Dict, Any, Type, Set

try:
    from django.utils.decorators import classproperty
except ImportError:
    from django.utils.functional import classproperty

from lucyfer.searchset.fields import BaseSearchField
from lucyfer.searchset.storage import SearchSetStorage
from lucyfer.searchset.utils import FieldType
from lucyfer.settings import lucyfer_settings


class BaseMetaClass:
    model = None

    show_suggestions = True
    escape_quotes_in_suggestions = True

    fields_to_exclude_from_mapping: List[str] = None
    fields_to_exclude_from_suggestions: List[str] = None

    search_fields_for_default_search: List[str] = None


class BaseSearchSetMetaClass(type):
    def __new__(mcs, name, bases, attrs):

        meta = mcs.get_meta(meta=attrs.pop("Meta", None))
        meta.search_fields_for_default_search = meta.search_fields_for_default_search or []

        searchset = super().__new__(mcs, name, bases, attrs)

        mcs.process_required_field(searchset=searchset, bases=bases)

        storage = mcs.get_storage(searchset=searchset, meta=meta, name=name, attrs=attrs, bases=bases)

        setattr(meta, "_storage", storage)
        setattr(searchset, "_meta", meta)

        return searchset

    _required_field = ["_field_base_class", "_default_field"]
    _default_reqiured_value = BaseSearchField

    @classmethod
    def process_required_field(mcs, searchset, bases):
        for name in mcs._required_field:
            if not hasattr(searchset, name) or getattr(searchset, name) == mcs._default_reqiured_value:
                value = getattr(bases[-1], name, mcs._default_reqiured_value) if bases else mcs._default_reqiured_value
                setattr(searchset, name, value)

    @classmethod
    def get_meta(mcs, meta) -> Type[BaseMetaClass]:
        if meta:
            class CurrentMeta(meta, BaseMetaClass):
                pass
        else:
            class CurrentMeta(BaseMetaClass):
                pass

        return CurrentMeta

    @classmethod
    def get_storage(mcs, searchset, meta, name, attrs, bases):
        field_name_to_field = mcs.get_field_name_to_field(base_field_class=searchset._field_base_class, attrs=attrs)
        mcs.validate_field_name_to_field(field_name_to_field=field_name_to_field, searchset_name=name)

        for field_name, field in field_name_to_field.items():
            field._contribute_to_class(field_name=field_name)

        fields_to_exclude_from_mapping = mcs.get_fields_to_exclude_from_mapping(
            searchset_fields_to_exclude_from_mapping=meta.fields_to_exclude_from_mapping or [],
            field_name_to_field=field_name_to_field,
        )

        fields_to_exclude_from_suggestions = mcs.get_fields_to_exclude_from_suggestions(
            searchset_fields_to_exclude_from_suggestions=meta.fields_to_exclude_from_suggestions or [],
            field_name_to_field=field_name_to_field,
        )

        field_class_for_default_searching = None
        if meta.search_fields_for_default_search:
            field_class_for_default_searching = searchset._field_class_for_default_searching(
                sources=meta.search_fields_for_default_search,
                use_field_class_for_sources=False
            )

        storage = SearchSetStorage(
            field_name_to_field=field_name_to_field,
            searchset_class=searchset,
            fields_to_exclude_from_mapping=fields_to_exclude_from_mapping,
            fields_to_exclude_from_suggestions=fields_to_exclude_from_suggestions,
            field_class_for_default_searching=field_class_for_default_searching,
        )

        bases_meta_classes = [base for base in bases
                              if hasattr(base, "_meta") and issubclass(base._meta, BaseMetaClass)]

        if bases_meta_classes:
            for base in bases_meta_classes:
                storage.field_name_to_field.update(
                    {
                        name: field if field.__class__ != base._default_field else searchset._default_field()
                        for name, field in base._meta._storage.field_name_to_field.items()
                    }
                )

        return storage

    @classmethod
    def get_field_name_to_field(mcs,
                                base_field_class: Type[BaseSearchField],
                                attrs: Dict[str, Any],
                                ) -> Dict[str, BaseSearchField]:
        """
        Returns dictionary with fields defined in searchset (field name: field instance)

        :param base_field_class: all fields must be inheritors of that class
        :param attrs: searchset class attributes
        """
        return {name: instance for name, instance in attrs.items() if isinstance(instance, base_field_class)}

    @classmethod
    def validate_field_name_to_field(mcs,
                                     field_name_to_field: Dict[str, BaseSearchField],
                                     searchset_name: str) -> None:
        for name, field in field_name_to_field.items():
            if field.sources and len(field.sources) == 1:
                assert field.sources[0] != name, \
                    f"Defining field source equals to field name in searchset doesn't make sense. " \
                    f"You have to remove sources from field '{name}' in {searchset_name}"

    @classmethod
    def get_fields_to_exclude_from_mapping(mcs,
                                           searchset_fields_to_exclude_from_mapping: List[str],
                                           field_name_to_field: Dict[str, BaseSearchField]) -> Set[str]:
        for name, field in field_name_to_field.items():
            if field.exclude_sources_from_mapping:
                searchset_fields_to_exclude_from_mapping.extend(field.sources)
        return set(searchset_fields_to_exclude_from_mapping)

    @classmethod
    def get_fields_to_exclude_from_suggestions(mcs,
                                               searchset_fields_to_exclude_from_suggestions: List[str],
                                               field_name_to_field: Dict[str, BaseSearchField]) -> Set[str]:
        searchset_fields_to_exclude_from_suggestions.extend(
            [name for name, field in field_name_to_field.items() if not field.show_suggestions]
        )
        return set(searchset_fields_to_exclude_from_suggestions)


class BaseSearchSet(metaclass=BaseSearchSetMetaClass):
    _field_base_class = BaseSearchField

    # default field uses for creating query for fields not defined in searchset class
    _default_field = BaseSearchField

    # field for default searching (without lucene-like syntax usage like `?search=ululu`)
    _field_class_for_default_searching = BaseSearchField

    # provides possibility to use auto cast for boolean/integer/etc fields by field classes usage
    # that means we analyze elastic mapping data types or django models to match it to field classes
    # TODO property
    _field_type_to_field_class: Optional[Dict[int, _field_base_class]] = None
    _raw_type_to_field_type: Optional[Dict[Any, int]] = None

    @classproperty
    def storage(cls):
        return cls._meta._storage

    @classmethod
    def get_fields_values(
            cls,
            qs,
            field_name,
            prefix='',
            cache_key="DEFAULT_KEY",
            max_return_suggestions_count=lucyfer_settings.CACHE_MAX_VALUES_COUNT_FOR_ONE_PREFIX,
            allow_empty_values=lucyfer_settings.ALLOW_EMPTY_SUGGESTIONS,
            sort_values=True,
    ) -> List[str]:
        """
        Returns search helpers for field by prefix.

        :param qs: queryset with additional filters or els search
        :param field_name: field name for get values
        :param prefix: prefix for searching values
        :param cache_key: additional argument to use possibility to cache different values for different users for ex.
        :param max_return_suggestions_count: by default is 10 but if you want more - you can change it
        :param allow_empty_values: if False you will miss empty values
        :params sort_values: sort option

        :return: list of values
        """
        if not cls._meta.show_suggestions:
            return list()

        field = cls.storage.field_source_to_field.get(field_name, cls._default_field())

        return field.get_values(
            qs=qs,
            prefix=prefix,
            cache_key=cache_key,
            model_name=cls._meta.model.__name__,
            escape_quotes_in_suggestions=cls._meta.escape_quotes_in_suggestions,
            max_return_suggestions_count=max_return_suggestions_count,
            allow_empty_values=allow_empty_values,
            sort_values=sort_values,
        )

    @classmethod
    def _get_raw_mapping(cls) -> Dict[str, FieldType]:
        """
        That method allows to get mapping in raw format. It have to be reimplemented
        """
        raise NotImplementedError()

    @classmethod
    def get_query_for_field(cls, condition):
        """
        Returns Q object with query for parsed condition
        """
        field = cls.storage.field_source_to_field.get(condition.name, cls._default_field())
        return field.get_query(condition)



from itertools import islice
from typing import List

from django.core.cache import cache

from lucyfer.searchset.fields.mapping.utils import escape_quotes, ignore_empty_values, custom_sorted
from lucyfer.settings import lucyfer_settings


class MappingMixin:
    """
    Class provides interface to get values-suggestions for field by some prefix with cache
    """

    def get_values(self,
                   qs,
                   model_name: str,
                   escape_quotes_in_suggestions: bool,
                   prefix: str = '',
                   cache_key: str = "DEFAULT_KEY",
                   max_return_suggestions_count: int = lucyfer_settings.CACHE_MAX_VALUES_COUNT_FOR_ONE_PREFIX,
                   allow_empty_values: bool = lucyfer_settings.ALLOW_EMPTY_SUGGESTIONS,
                   sort_values: bool = True) -> List[str]:
        if not self.show_suggestions:
            return list()

        return_sorted_if_need = lambda x: custom_sorted(x) if sort_values else x

        key = f"LUCYFER__{model_name}__{cache_key}__{prefix}"

        # possibly the values has been cached already
        if self._is_prefix_may_be_cached(prefix=prefix) and cache.get(key):
            return return_sorted_if_need(cache.get(key))

        values = self._get_values(qs=qs, prefix=prefix,
                                  escape_quotes_in_suggestions=escape_quotes_in_suggestions,
                                  allow_empty_values=allow_empty_values)

        # if None we will return ALL found values, else needed count
        if max_return_suggestions_count is not None:
            values = list(islice(values, max_return_suggestions_count))

        # next we have to cache it
        if self._is_prefix_may_be_cached(prefix=prefix):
            cache.set(key, values, lucyfer_settings.CACHE_TIME)

        return return_sorted_if_need(values)

    def prepare_qs_for_suggestions(self, qs, prefix: str):
        """
        Process qs or search. Add aggregation things to use it in later for values getting
        """
        raise NotImplementedError()

    def get_suggestions_from_prepared_qs(self, qs, prefix: str):
        """
        Execute the qs or search and get raw values from it
        """
        raise NotImplementedError()

    def _get_values(self,
                    qs,
                    prefix: str,
                    escape_quotes_in_suggestions: bool,
                    allow_empty_values: bool) -> List[str]:
        """
        Returns all possible values (NOT SLICED)
        """
        method = self.get_available_values_method()

        if callable(method):
            prefix_lower = prefix.lower()
            available_values = method(**self._available_values_method_kwargs)
            values = list(set((v for v in available_values if prefix_lower in v.lower())))
        else:
            qs = self.prepare_qs_for_suggestions(qs=qs, prefix=prefix)
            values = self.get_suggestions_from_prepared_qs(qs=qs, prefix=prefix)

        if not allow_empty_values:
            values = ignore_empty_values(values)

        if escape_quotes_in_suggestions:
            values = escape_quotes(values)

        return values

    def _is_prefix_may_be_cached(self, prefix):
        return lucyfer_settings.CACHE_SEARCH_VALUES and \
               self.use_cache_for_suggestions and \
               len(prefix) >= lucyfer_settings.CACHE_VALUES_MIN_LENGTH



from collections import defaultdict
from typing import List


from elasticsearch.exceptions import RequestError
from lucyfer.searchset.fields.mapping.base import MappingMixin


class ElasticMappingMixin(MappingMixin):
    def prepare_qs_for_suggestions(self, qs, prefix: str):
        search = qs.extra(size=0).query('query_string', **{"fields": self.sources, "query": f'*{prefix}*'})
        for source in self.sources:
            search.aggs.bucket(source, "terms", field=source)
        return search

    def get_suggestions_from_prepared_qs(self, qs, prefix: str) -> List[str]:
        try:
            aggs = qs.execute().to_dict().get("aggregations", defaultdict(dict))
        except RequestError:
            return []
        result = []
        for source in self.sources:
            if aggs[source].get("buckets"):
                result.extend(
                    [val for val
                     in [str(val["key"]) for val in aggs[source]["buckets"]]
                     if prefix in val]
                )
        return list(set(result))



from typing import Dict

from lucyfer.parser import LuceneToElasticParserMixin
from lucyfer.searchset.base import BaseSearchSet
from lucyfer.searchset.fields.elastic import default_elastic_field_types_to_fields, ElasticSearchField, \
    ElasticQueryStringField
from lucyfer.searchset.utils import FieldType
from lucyfer.utils import LuceneSearchException


elastic_data_type_to_field_type = {
    "long": FieldType.INTEGER,
    "integer": FieldType.INTEGER,
    "short": FieldType.INTEGER,
    "double": FieldType.FLOAT,
    "float": FieldType.FLOAT,
    "boolean": FieldType.BOOLEAN,
    "date": FieldType.TIMESTAMP,
    "date_nanos": FieldType.TIMESTAMP,
}


class ElasticSearchSet(LuceneToElasticParserMixin, BaseSearchSet):
    _field_base_class = ElasticSearchField
    _default_field = ElasticSearchField
    _field_class_for_default_searching = ElasticQueryStringField

    _field_type_to_field_class = default_elastic_field_types_to_fields
    _raw_type_to_field_type = elastic_data_type_to_field_type

    @classmethod
    def get_es_client(cls, **kwargs):
        raise NotImplementedError()

    @classmethod
    def filter(cls, search, search_terms, raise_exception=False):
        query = cls.parse(raw_expression=search_terms)
        if query is None:
            if raise_exception:
                raise LuceneSearchException()
            else:
                return None

        return search.query(query)

    @classmethod
    def _format_mapping_values(cls, mapping, prefix="") -> Dict[str, FieldType]:
        field_name_to_field_type = dict()

        for key, value in mapping.items():
            if "properties" in value:
                field_name_to_field_type.update(
                    cls._format_mapping_values(value["properties"], ".".join([prefix, key]) if prefix else key)
                )
            else:
                field_name = ".".join([prefix, key]) if prefix else key
                field_name_to_field_type[field_name] = cls._raw_type_to_field_type.get(value.get("type"))

        return field_name_to_field_type

    @classmethod
    def _get_raw_mapping(cls) -> Dict[str, FieldType]:
        model_instance = cls._meta.model()
        index_to_mapping = cls.get_es_client().indices.get_mapping(index=model_instance._get_index())
        if not index_to_mapping:
            return dict()

        last_index = max(index_to_mapping)

        try:
            last_mapping = index_to_mapping[last_index]['mappings'][model_instance._doc_type.name]['properties']
        except KeyError:
            return dict()

        field_name_to_field_type = cls._format_mapping_values(last_mapping)
        return field_name_to_field_type



import warnings
from dataclasses import dataclass
from typing import Dict, Any, Set, Optional

from lucyfer.searchset.fields import BaseSearchField, FieldType
from lucyfer.settings import lucyfer_settings


@dataclass
class SearchSetStorage:
    """
    Class provides availability to use fields in SearchSet class
    """

    searchset_class: Any

    # fields defined in searchset class and those instances
    field_name_to_field: Dict[str, BaseSearchField]

    fields_to_exclude_from_mapping: Set[str]
    fields_to_exclude_from_suggestions: Set[str]

    field_class_for_default_searching: Optional[BaseSearchField]

    _full_mapping = None

    @property
    def mapping(self):
        """
        Returns mapping for current searchset. Its looks like {field name: field}
        That property only contains fields not excluded from original mapping
        If you want to get values for some field in mapping - don't. You better use `field_source_to_field`
        """
        return {
            name: field for name, field in self.field_source_to_field.items()
            if name not in self.fields_to_exclude_from_mapping
        }

    @property
    def raw_mapping(self) -> Dict[str, FieldType]:
        """
        Caches raw mapping and return it
        """
        if not hasattr(self, 'raw_mapping_result'):
            result = self.searchset_class._get_raw_mapping()
            setattr(self, 'raw_mapping_result', result)
        return self.raw_mapping_result

    @property
    def field_source_to_field(self) -> Dict[str, BaseSearchField]:
        """
        Auto generated fields by type checking in raw mapping and sources handling in defined fields
        """
        if not hasattr(self, 'field_source_to_field_result'):
            # first process raw mapping
            source_to_field_from_raw_mapping = {
                name: self.searchset_class._field_type_to_field_class.get(
                    field_type, self.searchset_class._default_field
                )(show_suggestions=name not in self.fields_to_exclude_from_suggestions, sources=[name])

                for name, field_type in self.raw_mapping.items()
            }

            # then process defined fields and its sources
            source_to_field_from_user_fields = self.field_name_to_field.copy()
            source_to_field_from_user_fields_sources = {}

            # `missed_fields` uses for process possibly missed sources from case when defined
            # some field in searchset like this one:
            # x = SearchField(sources=["y"], use_field_class_for_sources=False)
            # and when we have not found "y" in raw mapping and we have no idea what field class we need to use.
            # it will be warning in the end of function.
            missed_fields = []

            for name, field in self.field_name_to_field.items():
                if not field.sources:
                    continue

                if not field.use_field_class_for_sources:
                    # we extend missed fields by all fields because after cycle we will filter it anyway
                    missed_fields.extend([source for source in field.sources])
                    continue

                source_to_field_from_user_fields_sources.update(
                    {
                        source: field.__class__(sources=[source],
                                                show_suggestions=source not in self.fields_to_exclude_from_suggestions,
                                                get_available_values_method=field._get_available_values_method,
                                                available_values_method_kwargs=field._available_values_method_kwargs,
                                                use_cache_for_suggestions=field.use_cache_for_suggestions)
                        for source in field.sources
                    }
                )

            # now result
            # we create an empty dict and update it by our dicts with order from low to high priority.
            # it means if user have wrote field "A" in searchset and we have found field "A" in raw mapping
            # priority of searchset is higher, so in result we will see users field, not field from raw mapping.
            result = {}
            result.update(source_to_field_from_raw_mapping)
            result.update(source_to_field_from_user_fields_sources)
            result.update(source_to_field_from_user_fields)

            # and check default searching field if presented
            if self.field_class_for_default_searching:
                result.update({lucyfer_settings.FIELD_NAME_FOR_DEFAULT_SEARCH: self.field_class_for_default_searching})

            missed_fields = [field for field in missed_fields if field not in result]
            if missed_fields:
                warnings.warn(f"There is some undefined fields in {self.searchset_class}: {', '.join(missed_fields)}")

            setattr(self, "field_source_to_field_result", result)
        return self.field_source_to_field_result



from typing import List

from django.db.models import QuerySet

from lucyfer.searchset.fields.mapping.base import MappingMixin


class DjangoMappingMixin(MappingMixin):
    def prepare_qs_for_suggestions(self, qs, prefix: str):
        if prefix:
            qss = [
                qs.filter(**{'{}__icontains'.format(source): prefix}).values_list(source, flat=True).distinct()
                for source in self.sources
            ]
        else:
            qss = [
                qs.values_list(source, flat=True).distinct()
                for source in self.sources
            ]

        if not qss:
            return []

        first = qss.pop()
        if not isinstance(first, QuerySet):
            return []

        return first.union(*qss)

    def get_suggestions_from_prepared_qs(self, qs, prefix: str) -> List[str]:
        return list(set(qs))



from typing import Dict, List, Optional, Callable, Any

from lucyparser.tree import Operator

from lucyfer.searchset.fields.mapping.base import MappingMixin
from lucyfer.settings import lucyfer_settings


class BaseSearchField(MappingMixin):
    """
    Base Field class for including in SearchSet classes
    """

    DEFAULT_LOOKUP: str
    OPERATOR_TO_LOOKUP: Dict[Operator, str] = dict()
    _default_get_available_values_method = None

    def __init__(self,
                 sources: Optional[List[str]] = None,
                 exclude_sources_from_mapping: bool = False,
                 show_suggestions: bool = True,
                 get_available_values_method: Optional[Callable[..., List[Any]]] = None,
                 available_values_method_kwargs: Optional[Dict[str, Any]] = None,
                 use_field_class_for_sources: bool = False,
                 use_cache_for_suggestions: bool = None,
                 *args, **kwargs):

        sources = list() if sources is None else sources
        self.sources = list(set(sources))

        self.exclude_sources_from_mapping = exclude_sources_from_mapping
        self.show_suggestions = show_suggestions
        self.use_field_class_for_sources = use_field_class_for_sources
        self._get_available_values_method = get_available_values_method
        self._available_values_method_kwargs = available_values_method_kwargs or {}

        if use_cache_for_suggestions is None:
            self.use_cache_for_suggestions = lucyfer_settings.CACHE_SEARCH_VALUES
        else:
            self.use_cache_for_suggestions = use_cache_for_suggestions

        for k, v in kwargs.items():
            assert not k in dir(self), f'Kwarg "{k}" clashes with another class attribute'
            setattr(self, k, v)

    def _contribute_to_class(self, field_name):
        """
        in searchset metaclass we have to contribute field name to field class if it was defined by user
        """
        self.sources = self.sources or [field_name]

    def cast_value(self, value: str):
        """
        Method for value casting if it necessary (or for ex. for search replaces)
        """
        return value

    def get_lookup(self, operator: Operator) -> str:
        """
        Returns lookup for searching
        """
        return self.OPERATOR_TO_LOOKUP.get(operator, self.DEFAULT_LOOKUP)

    def get_sources(self, field_name: str) -> List[str]:
        """
        Returns sources list
        Be careful: in the default field you MUST use that method instead of self.sources usage.
        That because you have no idea about field name you've got for query generation
        """
        return self.sources or [field_name]

    def get_query(self, condition):
        """
        Returns Q object with query
        """
        raise NotImplementedError()

    def match_all(self, value) -> bool:
        """
        Check filtration necessary
        """
        return value == "*"

    def get_available_values_method(self) -> Optional[Callable[..., List[Any]]]:
        return self._get_available_values_method or self._default_get_available_values_method


def negate_query_if_necessary(func):
    """
    Decorator to invert query if conditions operator was NEQ
    """
    def wrapper(self, condition):
        query = func(self, condition)

        if query is not None and condition.operator == Operator.NEQ:
            query = query.__invert__()

        return query
    return wrapper



from elasticsearch_dsl import Q
from lucyparser.tree import ExpressionNode, AndNode, OrNode, NotNode

from lucyfer.parser.base import BaseLuceneParserMixin
from lucyfer.settings import lucyfer_settings


class LuceneToElasticParserMixin(BaseLuceneParserMixin):
    @classmethod
    def _parse_tree(cls, tree):
        if isinstance(tree, ExpressionNode):
            if lucyfer_settings.SAVED_SEARCHES_ENABLE and tree.name == lucyfer_settings.SAVED_SEARCHES_KEY:
                return cls.get_saved_search(tree)

            return cls.get_query_for_field(tree) or ~Q()

        query = ~Q()

        is_and_node = isinstance(tree, AndNode)
        is_or_node = isinstance(tree, OrNode)
        is_not_node = isinstance(tree, NotNode)

        if is_and_node or is_not_node or is_or_node:
            queries = [cls._parse_tree(tree=child) for child in tree.children]

            query = queries[0]

            if is_and_node:
                for q in queries[1:]:
                    query = query & q

            elif is_or_node:
                for q in queries[1:]:
                    query = query | q

            elif is_not_node:
                for q in queries[1:]:
                    query = query & q
                query = ~Q(query)

        return query



from typing import List, Dict

from django.core.exceptions import FieldError
from django.db.models import ForeignKey, AutoField, BooleanField, BigAutoField, BigIntegerField, FloatField, \
    IntegerField, NullBooleanField, PositiveIntegerField, PositiveSmallIntegerField, DateTimeField, DateField

from lucyfer.searchset.base import BaseSearchSet
from lucyfer.searchset.fields.django import DjangoSearchField, DjangoSearchFieldWithoutWildcard, \
    default_django_field_types_to_fields
from lucyfer.searchset.utils import FieldType
from lucyfer.parser import LuceneToDjangoParserMixin
from lucyfer.utils import LuceneSearchException


django_model_field_to_field_type = {
    AutoField: FieldType.INTEGER,
    BigAutoField: FieldType.INTEGER,
    BigIntegerField: FieldType.INTEGER,
    BooleanField: FieldType.BOOLEAN,
    FloatField: FieldType.FLOAT,
    IntegerField: FieldType.INTEGER,
    NullBooleanField: FieldType.NULL_BOOLEAN,
    PositiveIntegerField: FieldType.INTEGER,  # todo positive integer field
    PositiveSmallIntegerField: FieldType.INTEGER,
    DateTimeField: FieldType.TIMESTAMP,
    DateField: FieldType.TIMESTAMP,
}


class DjangoSearchSet(LuceneToDjangoParserMixin, BaseSearchSet):
    _field_base_class = DjangoSearchFieldWithoutWildcard
    _default_field = DjangoSearchField
    _field_class_for_default_searching = DjangoSearchField

    _field_type_to_field_class = default_django_field_types_to_fields
    _raw_type_to_field_type = django_model_field_to_field_type

    @classmethod
    def filter(cls, queryset, search_terms, raise_exception=False):
        query = cls.parse(raw_expression=search_terms)
        try:
            return queryset.filter(query)
        except FieldError:
            if raise_exception:
                raise LuceneSearchException()

            return queryset.none()

    @classmethod
    def _get_raw_mapping(cls) -> Dict[str, FieldType]:
        return {field.name: cls._raw_type_to_field_type.get(field.__class__)
                for field in cls._meta.model._meta.fields
                if not isinstance(field, ForeignKey)}

    @classmethod
    def get_fields_sources(cls) -> List[str]:
        """
        Returns sources for field defined in searchset class
        """
        sources = []
        for name, field in cls.storage.field_name_to_field.items():
            sources.extend(field.sources)
        return sources
